# -*- coding: utf-8 -*-

import typing as T
import dataclasses

import polars as pl

from ..arg import REQ, NA, rm_na, T_KWARGS
from ..base_expr import ExprEnum, BaseExpr, expr_enum_to_klass_mapping, parse_expr
from ..utils_expr import (
    batch_to_jsonpolars_into_exprs,
    batch_to_jsonpolars_named_into_exprs,
    batch_to_polars_into_exprs,
    batch_to_polars_named_into_exprs,
    to_jsonpolars_other_expr,
    to_polars_other_expr,
)

if T.TYPE_CHECKING:  # pragma: no cover
    from .api import T_EXPR
    from ..typehint import IntoExpr


def ensure_struct(expr: "T_EXPR") -> pl.Expr:
    if isinstance(expr, Struct):
        return expr.to_polars()
    else:
        return expr.to_polars().struct


@dataclasses.dataclass
class Struct(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/struct.html
    """

    type: str = dataclasses.field(default=ExprEnum.struct.value)
    expr: "T_EXPR" = dataclasses.field(default=REQ)

    @classmethod
    def from_dict(cls, dct: T_KWARGS):
        return cls(
            expr=parse_expr(dct["expr"]),
        )

    def to_polars(self) -> pl.Expr:
        return ensure_struct(self.expr)


expr_enum_to_klass_mapping[ExprEnum.struct.value] = Struct


@dataclasses.dataclass
class StructField(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.struct.field.html
    """

    type: str = dataclasses.field(default=ExprEnum.struct_field.value)
    expr: T.Optional["T_EXPR"] = dataclasses.field(default=None)
    name: T.Union[str, T.List[str]] = dataclasses.field(default=REQ)
    more_names: T.List[str] = dataclasses.field(default_factory=list)

    @classmethod
    def from_dict(cls, dct: T_KWARGS):
        kwargs = dict(
            name=dct["name"],
            more_names=dct.get("more_names", []),
        )
        expr = dct["expr"]
        if expr is not None:
            kwargs["expr"] = to_jsonpolars_other_expr(dct["expr"])
        return cls(**kwargs)

    def to_polars(self) -> pl.Expr:
        if self.expr is None:
            if isinstance(self.name, str):
                return pl.field(self.name, *self.more_names)
            else:  # pragma: no cover
                raise TypeError("name must be a string if expr is None")
        else:
            expr = ensure_struct(self.expr)

        return expr.field(self.name, *self.more_names)


expr_enum_to_klass_mapping[ExprEnum.struct_field.value] = StructField


@dataclasses.dataclass
class StructRenameFields(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.struct.rename_fields.html
    """

    type: str = dataclasses.field(default=ExprEnum.struct_rename_fields.value)
    expr: "T_EXPR" = dataclasses.field(default=REQ)
    names: T.List[str] = dataclasses.field(default=REQ)

    @classmethod
    def from_dict(cls, dct: T_KWARGS):
        return cls(
            expr=parse_expr(dct["expr"]),
            names=dct["names"],
        )

    def to_polars(self) -> pl.Expr:
        return ensure_struct(self.expr).rename_fields(names=self.names)


expr_enum_to_klass_mapping[ExprEnum.struct_rename_fields.value] = StructRenameFields


@dataclasses.dataclass
class StructWithFields(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.struct.with_fields.html
    """

    type: str = dataclasses.field(default=ExprEnum.struct_with_fields.value)
    expr: "T_EXPR" = dataclasses.field(default=REQ)
    exprs: T.List["IntoExpr"] = dataclasses.field(default_factory=list)
    named_exprs: T.Dict[str, "IntoExpr"] = dataclasses.field(default_factory=dict)

    @classmethod
    def from_dict(cls, dct: T_KWARGS):
        return cls(
            expr=parse_expr(dct["expr"]),
            exprs=batch_to_jsonpolars_into_exprs(dct.get("exprs", list())),
            named_exprs=batch_to_jsonpolars_named_into_exprs(
                dct.get("named_exprs", dict())
            ),
        )

    def to_polars(self) -> pl.Expr:
        return ensure_struct(self.expr).with_fields(
            *batch_to_polars_into_exprs(self.exprs),
            **batch_to_polars_named_into_exprs(self.named_exprs),
        )


expr_enum_to_klass_mapping[ExprEnum.struct_with_fields.value] = StructWithFields



# -*- coding: utf-8 -*-

"""
Base DataFrame operation.
"""

import typing as T
import enum
import dataclasses

import polars as pl
from .model import BaseModel
from .base_expr import BaseExpr

from .exc import ParamError
from .arg import REQ, rm_na, T_KWARGS

if T.TYPE_CHECKING:  # pragma: no cover
    from .dfop.api import T_DFOP


class DfopEnum(str, enum.Enum):
    """ """

    # Aggregation
    count = "count"
    max = "max"
    max_horizontal = "max_horizontal"
    mean = "mean"
    mean_horizontal = "mean_horizontal"
    median = "median"
    min = "min"
    min_horizontal = "min_horizontal"
    product = "product"
    quantile = "quantile"
    std = "std"
    sum = "sum"
    sum_horizontal = "sum_horizontal"
    var = "var"
    # Attributes
    # Computation
    # Descriptive
    # Export
    # GroupBy
    group_by_agg = "group_by_agg"
    group_by_all = "group_by_all"
    group_by_count = "group_by_count"
    group_by_first = "group_by_first"
    group_by_head = "group_by_head"
    group_by_last = "group_by_last"
    group_by_len = "group_by_len"
    group_by_map_groups = "group_by_map_groups"
    group_by_max = "group_by_max"
    group_by_mean = "group_by_mean"
    group_by_median = "group_by_median"
    group_by_min = "group_by_min"
    group_by_n_unique = "group_by_n_unique"
    group_by_quantile = "group_by_quantile"
    group_by_sum = "group_by_sum"
    group_by_tail = "group_by_tail"
    # Manipulation / selection
    bottom_k = "bottom_k"
    cast = "cast"
    clear = "clear"
    clone = "clone"
    drop = "drop"
    drop_in_place = "drop_in_place"
    drop_nulls = "drop_nulls"
    explode = "explode"
    extend = "extend"
    fill_nan = "fill_nan"
    fill_null = "fill_null"
    filter = "filter"
    gather_every = "gather_every"
    get_column = "get_column"
    get_column_index = "get_column_index"
    get_columns = "get_columns"
    group_by = "group_by"
    group_by_dynamic = "group_by_dynamic"
    head = "head"
    hstack = "hstack"
    insert_column = "insert_column"
    interpolate = "interpolate"
    item = "item"
    iter_columns = "iter_columns"
    iter_rows = "iter_rows"
    iter_slices = "iter_slices"
    join = "join"
    join_asof = "join_asof"
    limit = "limit"
    melt = "melt"
    merge_sorted = "merge_sorted"
    partition_by = "partition_by"
    pipe = "pipe"
    pivot = "pivot"
    rechunk = "rechunk"
    rename = "rename"
    replace_column = "replace_column"
    reverse = "reverse"
    rolling = "rolling"
    row = "row"
    rows = "rows"
    rows_by_key = "rows_by_key"
    sample = "sample"
    select = "select"
    select_seq = "select_seq"
    set_sorted = "set_sorted"
    shift = "shift"
    shrink_to_fit = "shrink_to_fit"
    slice = "slice"
    sort = "sort"
    sql = "sql"
    tail = "tail"
    to_dummies = "to_dummies"
    to_series = "to_series"
    top_k = "top_k"
    transpose = "transpose"
    unique = "unique"
    unnest = "unnest"
    unpivot = "unpivot"
    unstack = "unstack"
    update = "update"
    upsample = "upsample"
    vstack = "vstack"
    with_columns = "with_columns"
    with_columns_seq = "with_columns_seq"
    with_row_count = "with_row_count"
    with_row_index = "with_row_index"
    # Miscellaneous
    # Plot
    # Style


def to_dict(inst) -> "T_KWARGS":
    """
    Convert an instance of ``BaseExpr`` to a dict. This dict can be used in
    ``BaseExpr.from_dict`` method to create a identical instance of the original
    ``BaseExpr`` instance.
    """
    if isinstance(inst, BaseExpr):
        return inst.to_dict()
    elif isinstance(inst, (tuple, list)):
        return type(inst)([to_dict(v) for v in inst])
    elif isinstance(inst, dict):
        kwargs = {k: to_dict(v) for k, v in inst.items()}
        return rm_na(**kwargs)
    else:
        return inst


@dataclasses.dataclass
class BaseDfop(BaseModel):
    type: str = dataclasses.field(default=REQ)

    def _validate(self):
        for field in dataclasses.fields(self.__class__):
            if field.init:
                k = field.name
                if getattr(self, k) is REQ:  # pragma: no cover
                    raise ParamError(f"Field {k!r} is required for {self.__class__}.")

    def __post_init__(self):
        self._validate()

    def to_dict(self) -> T_KWARGS:
        kwargs = dict()
        for field in dataclasses.fields(self.__class__):
            value = getattr(self, field.name)
            kwargs[field.name] = to_dict(value)
        return rm_na(**kwargs)

    @classmethod
    def from_dict(cls, dct: T_KWARGS):
        """
        Create an instance of ``BaseExpr`` from either a human created dict,
        or a dict created by the ``BaseExpr.to_dict`` method.
        """
        req_kwargs, opt_kwargs = cls._split_req_opt(dct)
        return cls(**req_kwargs, **rm_na(**opt_kwargs))

    def to_polars(self, df: pl.DataFrame) -> pl.DataFrame:
        raise NotImplementedError()


dfop_enum_to_klass_mapping: T.Dict[str, T.Type["T_DFOP"]] = dict()


def parse_dfop(dct: T.Dict[str, T.Any]) -> "T_DFOP":
    """
    Note: you have to import everything in the :mod:`jsonpolars.dfop` module
    to make this work.
    """
    return dfop_enum_to_klass_mapping[dct["type"]].from_dict(dct)



# -*- coding: utf-8 -*-

import typing as T
from .arg import PRE


if T.TYPE_CHECKING:  # pragma: no cover
    from .expr.api import T_EXPR


_ = PRE


def chain(*args: "T_EXPR") -> "T_EXPR":
    ex = None
    for arg in args:
        # print(ex, arg)
        if ex is not None:
            setattr(arg, "expr", ex)
        ex = arg
    # print(ex)
    return ex



# -*- coding: utf-8 -*-

from jsonpolars.expr import api as expr
from jsonpolars.chain import chain, PRE
from jsonpolars.tests.expr_case import Case


data = {
    "id": 1,
    "name": "Alice",
    "profile": {
        "ssn": "1234567890",
        "passport": "E1234567",
    },
}
case_field_1 = Case(
    input_records=[
        {"data": data},
    ],
    expr=expr.Alias(
        expr=expr.StructField(
            expr=expr.Struct(
                expr=expr.Column(name="data"),
            ),
            name="id",
        ),
        name="res",
    ),
    expected_output_records=[
        {"data": data, "res": 1},
    ],
)
case_field_2 = Case(
    input_records=[
        {"data": data},
    ],
    expr=expr.Alias(
        expr=expr.StructField(
            expr=expr.Column(name="data"),
            name="id",
        ),
        name="res",
    ),
    expected_output_records=[
        {"data": data, "res": 1},
    ],
)
case_field_3 = Case(
    input_records=[
        {"data": data},
    ],
    expr=expr.StructField(
        expr=expr.Column(name="data"),
        name="*",
    ),
    expected_output_records=[
        {
            "data": data,
            "id": 1,
            "name": "Alice",
            "profile": {
                "ssn": "1234567890",
                "passport": "E1234567",
            },
        },
    ],
)
case_field_4 = Case(
    input_records=[
        {"data": data},
    ],
    expr=expr.StructField(
        expr=expr.Column(name="data"),
        name=["id", "name"],
    ),
    expected_output_records=[
        {
            "data": data,
            "id": 1,
            "name": "Alice",
        },
    ],
)
case_field_5 = Case(
    input_records=[
        {"data": data},
    ],
    expr=expr.StructField(
        expr=expr.Column(name="data"),
        name="id",
        more_names=["name"],
    ),
    expected_output_records=[
        {
            "data": data,
            "id": 1,
            "name": "Alice",
        },
    ],
)
case_field_6 = Case(
    input_records=[
        {"a": {"b": {"c": 1}}},
    ],
    expr=chain(
        expr.Column(name="a"),
        expr.StructField(expr=PRE, name="b"),
        expr.Alias(expr=PRE, name="d"),
    ),
    expected_output_records=[
        {"a": {"b": {"c": 1}}, "d": 1},
    ],
)
case_rename_fields_1 = Case(
    input_records=[
        {"data": {"id": 1, "name": "Alice"}},
    ],
    expr=expr.StructRenameFields(
        expr=expr.Column(name="data"),
        names=["user_id", "user_name"],
    ),
    expected_output_records=[
        {"data": {"user_id": 1, "user_name": "Alice"}},
    ],
)
case_rename_fields_2 = Case(
    input_records=[
        {"data": {"id": 1, "name": "Alice"}},
    ],
    expr=expr.StructRenameFields(
        expr=expr.Column(name="data"),
        # if number of item in names doesn't match number of fields,
        # it will follow the order of the fields and rename it with the names
        names=["user_name"],
    ),
    expected_output_records=[
        {"data": {"user_name": 1}},
    ],
)
case_with_fields_1 = Case(
    input_records=[
        {"data": {"a": 2, "b": 3}},
    ],
    expr=expr.StructWithFields(
        expr=expr.Column(name="data"),
        exprs=[
            expr.Alias(
                name="c",
                expr=expr.Plus(
                    left=expr.StructField(name="a"),
                    right=expr.StructField(name="b"),
                ),
            )
        ],
        named_exprs={
            "d": expr.Multiply(
                left=expr.StructField(name="a"),
                right=expr.StructField(name="b"),
            )
        },
    ),
    expected_output_records=[
        {"data": {"a": 2, "b": 3, "c": 5, "d": 6}},
    ],
)
case_with_fields_2 = Case(
    input_records=[
        {"data": {"a": 2, "b": 3}},
    ],
    expr=chain(
        expr.Column(name="data"),
        expr.StructWithFields(
            expr=PRE,
            exprs=[
                chain(
                    expr.Plus(
                        left=expr.StructField(name="a"),
                        right=expr.StructField(name="b"),
                    ),
                    expr.Alias(expr=PRE, name="c"),
                )
            ],
            named_exprs={
                "d": expr.Multiply(
                    left=expr.StructField(name="a"), right=expr.StructField(name="b")
                ),
            },
        ),
    ),
    expected_output_records=[
        {"data": {"a": 2, "b": 3, "c": 5, "d": 6}},
    ],
)


def test():
    print("")

    case_field_1.run_with_columns_test()
    case_field_2.run_with_columns_test()
    case_field_3.run_with_columns_test()
    case_field_4.run_with_columns_test()
    case_field_5.run_with_columns_test()
    case_rename_fields_1.run_with_columns_test()
    case_rename_fields_2.run_with_columns_test()
    case_with_fields_1.run_with_columns_test()
    case_with_fields_2.run_with_columns_test()


if __name__ == "__main__":
    from jsonpolars.tests import run_cov_test

    run_cov_test(__file__, "jsonpolars.expr.struct", preview=False)



# -*- coding: utf-8 -*-

import typing as T
import dataclasses


@dataclasses.dataclass(frozen=True)
class _REQUIRED:
    def __eq__(self, other):
        return isinstance(other, _REQUIRED)


@dataclasses.dataclass(frozen=True)
class _NOTHING:
    def __eq__(self, other):
        print(self, other)
        return isinstance(other, _NOTHING)


@dataclasses.dataclass(frozen=True)
class _PRE:
    def __eq__(self, other):
        print(self, other)
        return isinstance(other, PRE)


REQ = _REQUIRED()
NA = _NOTHING()
PRE = _PRE()

T_KWARGS = T.Dict[str, T.Any]


def rm_na(**kwargs) -> T_KWARGS:
    """
    Remove NA values from kwargs.
    """
    return {
        key: value
        for key, value in kwargs.items()
        if isinstance(value, _NOTHING) is False
    }



# -*- coding: utf-8 -*-

"""
This test case also an example to show how to implement the dataclass model correctly.
"""

import typing as T
import dataclasses
from datetime import datetime

import pytest

from jsonpolars.exc import ParamError
from jsonpolars.arg import REQ, NA, rm_na, T_KWARGS
from jsonpolars.base_expr import BaseExpr, parse_expr, expr_enum_to_klass_mapping


@dataclasses.dataclass
class Person(BaseExpr):
    type: str = dataclasses.field(default="test_person")
    name: str = dataclasses.field(default=REQ)
    age: int = dataclasses.field(default=20)
    gender: str = dataclasses.field(default=NA)


expr_enum_to_klass_mapping["test_person"] = Person


@dataclasses.dataclass
class Model(BaseExpr):
    type: str = dataclasses.field(default="test_model")
    id: int = dataclasses.field(default=REQ)
    name: str = dataclasses.field(default="my_model")
    create_at: str = dataclasses.field(default=NA)
    person: Person = dataclasses.field(default=REQ)
    person_list: T.List[Person] = dataclasses.field(default_factory=list)
    person_dict: T.Dict[str, Person] = dataclasses.field(default_factory=dict)

    @classmethod
    def from_dict(cls, dct: T_KWARGS):
        req_kwargs, opt_kwargs = cls._split_req_opt(dct)
        req_kwargs["person"] = parse_expr(req_kwargs["person"])
        opt_kwargs["person_list"] = [
            parse_expr(d) for d in opt_kwargs.get("person_list", [])
        ]
        opt_kwargs["person_dict"] = {
            k: parse_expr(v) for k, v in opt_kwargs.get("person_dict", {}).items()
        }
        return cls(**req_kwargs, **rm_na(**opt_kwargs))


expr_enum_to_klass_mapping["test_model"] = Model


@dataclasses.dataclass
class Record(BaseExpr):
    type: str = dataclasses.field(default="test_record")
    create_time: datetime = dataclasses.field(default=REQ)

    def to_dict(self) -> T_KWARGS:
        dct = super().to_dict()
        dct["create_time"] = dct["create_time"].isoformat()
        return dct

    @classmethod
    def from_dict(cls, dct: T_KWARGS):
        return cls(create_time=datetime.fromisoformat(dct["create_time"]))


expr_enum_to_klass_mapping["test_record"] = Record


def test_init():
    # id field is required
    with pytest.raises(ParamError):
        model = Model()

    # person field is required
    with pytest.raises(ParamError):
        model = Model(id=1)


def test_to_dict_from_dict():
    model = Model(
        id=1,
        person=Person(name="Alice"),
        person_list=[Person(name="Bob")],
        person_dict={"c": Person(name="Cathy")},
    )
    assert model.create_at == NA
    dct = model.to_dict()
    assert dct == {
        "type": "test_model",
        "id": 1,
        "name": "my_model",
        "person": {"type": "test_person", "name": "Alice", "age": 20},
        "person_list": [{"type": "test_person", "name": "Bob", "age": 20}],
        "person_dict": {"c": {"type": "test_person", "name": "Cathy", "age": 20}},
    }

    model1 = Model.from_dict(dct)
    assert model1 == model

    req_kwargs, opt_kwargs = Model._split_req_opt(dct)
    assert req_kwargs == {
        "id": 1,
        "person": {"type": "test_person", "name": "Alice", "age": 20},
    }
    assert opt_kwargs == {
        "type": "test_model",
        "name": "my_model",
        "person_list": [{"type": "test_person", "name": "Bob", "age": 20}],
        "person_dict": {"c": {"type": "test_person", "name": "Cathy", "age": 20}},
    }

    with pytest.raises(ParamError):
        Model._split_req_opt({})


class TestRecord:
    def test_to_dict_from_dict(self):
        record = Record(create_time=datetime(2000, 1, 1))
        dct = record.to_dict()
        assert dct == {"type": "test_record", "create_time": "2000-01-01T00:00:00"}

        record1 = Record.from_dict(dct)
        assert record1 == record


if __name__ == "__main__":
    from jsonpolars.tests import run_cov_test

    run_cov_test(__file__, "jsonpolars.base_expr", preview=False)



# -*- coding: utf-8 -*-

import typing as T
import dataclasses

from .exc import ParamError
from .arg import _REQUIRED, REQ, rm_na, T_KWARGS


@dataclasses.dataclass
class BaseModel:
    def _validate(self):
        for field in dataclasses.fields(self.__class__):
            if field.init:
                k = field.name
                if getattr(self, k) is REQ:  # pragma: no cover
                    raise ParamError(f"Field {k!r} is required for {self.__class__}.")

    def __post_init__(self):
        self._validate()

    def to_dict(self) -> T_KWARGS:
        raise NotImplementedError

    @classmethod
    def from_dict(cls, dct: T_KWARGS):
        raise NotImplementedError

    @classmethod
    def _split_req_opt(cls, kwargs: T_KWARGS) -> T.Tuple[T_KWARGS, T_KWARGS]:
        req_kwargs, opt_kwargs = dict(), dict()
        for field in dataclasses.fields(cls):
            if isinstance(field.default, _REQUIRED):
                try:
                    req_kwargs[field.name] = kwargs[field.name]
                except KeyError:
                    raise ParamError(
                        f"{field.name!r} is a required parameter for {cls}!"
                    )
            else:
                try:
                    opt_kwargs[field.name] = kwargs[field.name]
                except KeyError:
                    pass
        opt_kwargs = rm_na(**opt_kwargs)
        return req_kwargs, opt_kwargs



# -*- coding: utf-8 -*-

import typing as T
import dataclasses

import polars as pl

from ..arg import REQ, NA, rm_na, T_KWARGS
from ..utils_expr import (
    batch_to_jsonpolars_into_exprs,
    batch_to_jsonpolars_named_into_exprs,
    batch_to_polars_into_exprs,
    batch_to_polars_named_into_exprs,
)
from ..base_dfop import DfopEnum, BaseDfop, dfop_enum_to_klass_mapping

if T.TYPE_CHECKING:  # pragma: no cover
    from .api import T_DFOP
    from ..expr.api import T_EXPR
    from ..typehint import IntoExpr, ColumnNameOrSelector


@dataclasses.dataclass
class Select(BaseDfop):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/dataframe/api/polars.DataFrame.select.html
    """

    type: str = dataclasses.field(default=DfopEnum.select.value)
    exprs: T.List["IntoExpr"] = dataclasses.field(default_factory=list)
    named_exprs: T.Dict[str, "IntoExpr"] = dataclasses.field(default_factory=dict)

    @classmethod
    def from_dict(cls, dct: T_KWARGS):
        return cls(
            exprs=batch_to_jsonpolars_into_exprs(dct.get("exprs", list())),
            named_exprs=batch_to_jsonpolars_named_into_exprs(
                dct.get("named_exprs", dict())
            ),
        )

    def to_polars(self, df: pl.DataFrame) -> pl.DataFrame:
        return df.select(
            *batch_to_polars_into_exprs(self.exprs),
            **batch_to_polars_named_into_exprs(self.named_exprs),
        )


dfop_enum_to_klass_mapping[DfopEnum.select.value] = Select


@dataclasses.dataclass
class Rename(BaseDfop):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/dataframe/api/polars.DataFrame.rename.html
    """

    type: str = dataclasses.field(default=DfopEnum.rename.value)
    mapping: T.Union[T.Dict[str, str], T.Callable[[str], str]] = dataclasses.field(
        default=REQ
    )

    def to_polars(self, df: pl.DataFrame) -> pl.DataFrame:
        return df.rename(self.mapping)


dfop_enum_to_klass_mapping[DfopEnum.rename.value] = Rename


@dataclasses.dataclass
class Drop(BaseDfop):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/dataframe/api/polars.DataFrame.drop.html
    """

    type: str = dataclasses.field(default=DfopEnum.drop.value)
    columns: T.List["ColumnNameOrSelector"] = dataclasses.field(default=REQ)
    strict: bool = dataclasses.field(default=NA)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(
            columns=batch_to_jsonpolars_into_exprs(dct["columns"]),
            **rm_na(
                strict=dct.get("strict", NA),
            ),
        )

    def to_polars(self, df: pl.DataFrame) -> pl.DataFrame:
        return df.drop(
            *batch_to_polars_into_exprs(self.columns),
            **rm_na(
                strict=self.strict,
            ),
        )


dfop_enum_to_klass_mapping[DfopEnum.drop.value] = Drop


@dataclasses.dataclass
class WithColumns(BaseDfop):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/dataframe/api/polars.DataFrame.with_columns.html
    """

    type: str = dataclasses.field(default=DfopEnum.with_columns.value)
    exprs: T.List["IntoExpr"] = dataclasses.field(default_factory=list)
    named_exprs: T.Dict[str, "IntoExpr"] = dataclasses.field(default_factory=dict)

    @classmethod
    def from_dict(cls, dct: T_KWARGS):
        return cls(
            exprs=batch_to_jsonpolars_into_exprs(dct.get("exprs", list())),
            named_exprs=batch_to_jsonpolars_named_into_exprs(
                dct.get("named_exprs", dict())
            ),
        )

    def to_polars(self, df: pl.DataFrame) -> pl.DataFrame:
        return df.with_columns(
            *batch_to_polars_into_exprs(self.exprs),
            **batch_to_polars_named_into_exprs(self.named_exprs),
        )


dfop_enum_to_klass_mapping[DfopEnum.with_columns.value] = WithColumns


@dataclasses.dataclass
class Head(BaseDfop):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/dataframe/api/polars.DataFrame.head.html
    """

    type: str = dataclasses.field(default=DfopEnum.head.value)
    n: int = dataclasses.field(default=NA)

    def to_polars(self, df: pl.DataFrame) -> pl.DataFrame:
        return df.head(**rm_na(n=self.n))


dfop_enum_to_klass_mapping[DfopEnum.head.value] = Head


@dataclasses.dataclass
class Tail(BaseDfop):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/dataframe/api/polars.DataFrame.tail.html
    """

    type: str = dataclasses.field(default=DfopEnum.tail.value)
    n: int = dataclasses.field(default=NA)

    def to_polars(self, df: pl.DataFrame) -> pl.DataFrame:
        return df.tail(**rm_na(n=self.n))


dfop_enum_to_klass_mapping[DfopEnum.tail.value] = Tail


@dataclasses.dataclass
class Sort(BaseDfop):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/dataframe/api/polars.DataFrame.sort.html
    """

    type: str = dataclasses.field(default=DfopEnum.sort.value)
    by: T.List["IntoExpr"] = dataclasses.field(default=REQ)
    descending: T.Union[bool, T.List[bool]] = dataclasses.field(default=NA)
    nulls_last: T.Union[bool, T.List[bool]] = dataclasses.field(default=NA)
    multithreaded: bool = dataclasses.field(default=NA)
    maintain_order: bool = dataclasses.field(default=NA)

    @classmethod
    def from_dict(cls, dct: T_KWARGS):
        req_kwargs, opt_kwargs = cls._split_req_opt(dct)
        req_kwargs["by"] = batch_to_jsonpolars_into_exprs(
            batch_to_jsonpolars_into_exprs(dct["by"])
        )
        return cls(**req_kwargs, **rm_na(**opt_kwargs))

    def to_polars(self, df: pl.DataFrame) -> pl.DataFrame:
        return df.sort(
            *batch_to_polars_into_exprs(self.by),
            **rm_na(
                descending=self.descending,
                nulls_last=self.nulls_last,
                multithreaded=self.multithreaded,
                maintain_order=self.maintain_order,
            ),
        )


dfop_enum_to_klass_mapping[DfopEnum.sort.value] = Sort


@dataclasses.dataclass
class DropNulls(BaseDfop):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/dataframe/api/polars.DataFrame.drop_nulls.html
    """

    type: str = dataclasses.field(default=DfopEnum.drop_nulls.value)
    subset: T.List["ColumnNameOrSelector"] = dataclasses.field(default=NA)

    @classmethod
    def from_dict(cls, dct: T_KWARGS):
        req_kwargs, opt_kwargs = cls._split_req_opt(dct)
        if "subset" in opt_kwargs:
            opt_kwargs["subset"] = batch_to_jsonpolars_into_exprs(opt_kwargs["subset"])
        return cls(**req_kwargs, **rm_na(**opt_kwargs))

    def to_polars(self, df: pl.DataFrame) -> pl.DataFrame:
        kwargs = dict()
        if isinstance(self.subset, list):
            kwargs["subset"] = batch_to_polars_into_exprs(self.subset)
        return df.drop_nulls(**kwargs)


dfop_enum_to_klass_mapping[DfopEnum.drop_nulls.value] = DropNulls



# -*- coding: utf-8 -*-

import typing as T
import dataclasses

import polars as pl

from ..arg import REQ, NA, rm_na, T_KWARGS
from ..base_expr import ExprEnum, BaseExpr, expr_enum_to_klass_mapping, parse_expr
from ..utils_expr import str_to_polars_dtype_mapping, polars_dtype_to_str_mapping

if T.TYPE_CHECKING:  # pragma: no cover
    from .api import T_EXPR


@dataclasses.dataclass
class Cast(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.cast.html
    """

    type: str = dataclasses.field(default=ExprEnum.cast.value)
    expr: "T_EXPR" = dataclasses.field(default=REQ)
    dtype: T.Union[str, "pl.DataType", T.Type["pl.DataType"]] = dataclasses.field(
        default=REQ
    )

    def to_dict(self) -> T_KWARGS:
        dct = super().to_dict()
        if "dtype" in dct:
            if isinstance(self.dtype, str):
                dtype = self.dtype
            else:
                try:
                    dtype = polars_dtype_to_str_mapping[type(self.dtype)]
                except KeyError:
                    dtype = polars_dtype_to_str_mapping[self.dtype]
            dct["dtype"] = dtype
        return dct

    @classmethod
    def from_dict(cls, dct: T_KWARGS):
        req_kwargs, opt_kwargs = cls._split_req_opt(dct)
        req_kwargs["expr"] = parse_expr(req_kwargs["expr"])
        return cls(**req_kwargs, **rm_na(**opt_kwargs))

    def to_polars(self) -> pl.Expr:
        if isinstance(self.dtype, str):
            dtype = str_to_polars_dtype_mapping[self.dtype]
        else:
            dtype = self.dtype
        return self.expr.to_polars().cast(dtype)


expr_enum_to_klass_mapping[ExprEnum.cast.value] = Cast



# -*- coding: utf-8 -*-

import typing as T
import dataclasses

import polars as pl

from ..arg import REQ, NA, rm_na, T_KWARGS
from ..base_expr import ExprEnum, BaseExpr, expr_enum_to_klass_mapping, parse_expr
from ..utils_expr import to_jsonpolars_other_expr, to_polars_other_expr

if T.TYPE_CHECKING:  # pragma: no cover
    from .api import T_EXPR


def ensure_string(expr: "T_EXPR") -> pl.Expr:
    if isinstance(expr, String):
        return expr.to_polars()
    else:
        return expr.to_polars().str


@dataclasses.dataclass
class String(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/string.html
    """

    type: str = dataclasses.field(default=ExprEnum.string.value)
    expr: "T_EXPR" = dataclasses.field(default=REQ)

    @classmethod
    def from_dict(cls, dct: T_KWARGS):
        return cls(
            expr=parse_expr(dct["expr"]),
        )

    def to_polars(self) -> pl.Expr:
        return ensure_string(self.expr)


expr_enum_to_klass_mapping[ExprEnum.string.value] = String


@dataclasses.dataclass
class Split(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.str.split.html
    """

    type: str = dataclasses.field(default=ExprEnum.str_split.value)
    expr: "T_EXPR" = dataclasses.field(default=REQ)
    by: str = dataclasses.field(default=REQ)
    inclusive: bool = dataclasses.field(default=NA)

    @classmethod
    def from_dict(cls, dct: T_KWARGS):
        req_kwargs, opt_kwargs = cls._split_req_opt(dct)
        req_kwargs["expr"] = parse_expr(req_kwargs["expr"])
        return cls(**req_kwargs, **rm_na(**opt_kwargs))

    def to_polars(self) -> pl.Expr:
        return ensure_string(self.expr).split(
            by=self.by,
            **rm_na(inclusive=self.inclusive),
        )


expr_enum_to_klass_mapping[ExprEnum.str_split.value] = Split


@dataclasses.dataclass
class StrJoin(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.str.join.html
    """

    type: str = dataclasses.field(default=ExprEnum.str_join.value)
    expr: "T_EXPR" = dataclasses.field(default=REQ)
    delimiter: str = dataclasses.field(default=NA)
    ignore_nulls: bool = dataclasses.field(default=NA)

    @classmethod
    def from_dict(cls, dct: T_KWARGS):
        req_kwargs, opt_kwargs = cls._split_req_opt(dct)
        req_kwargs["expr"] = parse_expr(req_kwargs["expr"])
        return cls(**req_kwargs, **rm_na(**opt_kwargs))

    def to_polars(self) -> pl.Expr:
        return ensure_string(self.expr).join(
            **rm_na(
                delimiter=self.delimiter,
                ignore_nulls=self.ignore_nulls,
            )
        )


expr_enum_to_klass_mapping[ExprEnum.str_join.value] = StrJoin


@dataclasses.dataclass
class StrContains(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.str.contains.html
    """

    type: str = dataclasses.field(default=ExprEnum.str_contains.value)
    expr: "T_EXPR" = dataclasses.field(default=REQ)
    pattern: T.Union[str, "T_EXPR"] = dataclasses.field(default=REQ)
    literal: bool = dataclasses.field(default=NA)
    strict: bool = dataclasses.field(default=NA)

    @classmethod
    def from_dict(cls, dct: T_KWARGS):
        req_kwargs, opt_kwargs = cls._split_req_opt(dct)
        req_kwargs["expr"] = parse_expr(req_kwargs["expr"])
        req_kwargs["pattern"] = to_jsonpolars_other_expr(req_kwargs["pattern"])
        return cls(**req_kwargs, **rm_na(**opt_kwargs))

    def to_polars(self) -> pl.Expr:
        return ensure_string(self.expr).contains(
            pattern=to_polars_other_expr(self.pattern),
            **rm_na(
                literal=self.literal,
                strict=self.strict,
            ),
        )


expr_enum_to_klass_mapping[ExprEnum.str_contains.value] = StrContains


@dataclasses.dataclass
class StrDecode(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.str.decode.html
    """

    type: str = dataclasses.field(default=ExprEnum.str_decode.value)
    expr: "T_EXPR" = dataclasses.field(default=REQ)
    encoding: str = dataclasses.field(default=REQ)
    strict: bool = dataclasses.field(default=NA)

    @classmethod
    def from_dict(cls, dct: T_KWARGS):
        req_kwargs, opt_kwargs = cls._split_req_opt(dct)
        req_kwargs["expr"] = parse_expr(req_kwargs["expr"])
        return cls(**req_kwargs, **rm_na(**opt_kwargs))

    def to_polars(self) -> pl.Expr:
        return ensure_string(self.expr).decode(
            encoding=self.encoding,
            **rm_na(
                strict=self.strict,
            ),
        )


expr_enum_to_klass_mapping[ExprEnum.str_decode.value] = StrDecode


@dataclasses.dataclass
class StrEncode(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.str.encode.html
    """

    type: str = dataclasses.field(default=ExprEnum.str_encode.value)
    expr: "T_EXPR" = dataclasses.field(default=REQ)
    encoding: str = dataclasses.field(default=REQ)

    @classmethod
    def from_dict(cls, dct: T_KWARGS):
        return cls(
            expr=parse_expr(dct["expr"]),
            encoding=dct["encoding"],
        )

    def to_polars(self) -> pl.Expr:
        return ensure_string(self.expr).encode(encoding=self.encoding)


expr_enum_to_klass_mapping[ExprEnum.str_encode.value] = StrEncode


@dataclasses.dataclass
class StrStartsWith(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.str.starts_with.html
    """

    type: str = dataclasses.field(default=ExprEnum.str_starts_with.value)
    expr: "T_EXPR" = dataclasses.field(default=REQ)
    prefix: T.Union[str, "T_EXPR"] = dataclasses.field(default=REQ)

    @classmethod
    def from_dict(cls, dct: T_KWARGS):
        return cls(
            expr=parse_expr(dct["expr"]),
            prefix=to_jsonpolars_other_expr(dct["prefix"]),
        )

    def to_polars(self) -> pl.Expr:
        return ensure_string(self.expr).starts_with(
            prefix=to_polars_other_expr(self.prefix)
        )


expr_enum_to_klass_mapping[ExprEnum.str_starts_with.value] = StrStartsWith


@dataclasses.dataclass
class StrEndsWith(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.str.ends_with.html
    """

    type: str = dataclasses.field(default=ExprEnum.str_ends_with.value)
    expr: "T_EXPR" = dataclasses.field(default=REQ)
    suffix: T.Union[str, "T_EXPR"] = dataclasses.field(default=REQ)

    @classmethod
    def from_dict(cls, dct: T_KWARGS):
        return cls(
            expr=parse_expr(dct["expr"]),
            suffix=to_jsonpolars_other_expr(dct["suffix"]),
        )

    def to_polars(self) -> pl.Expr:
        return ensure_string(self.expr).ends_with(
            suffix=to_polars_other_expr(self.suffix)
        )


expr_enum_to_klass_mapping[ExprEnum.str_ends_with.value] = StrEndsWith


@dataclasses.dataclass
class StrToDatetime(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.str.to_datetime.html
    """

    type: str = dataclasses.field(default=ExprEnum.str_to_datetime.value)
    expr: "T_EXPR" = dataclasses.field(default=REQ)
    format: T.Optional[str] = dataclasses.field(default=NA)
    time_unit: T.Optional[str] = dataclasses.field(default=NA)
    time_zone: T.Optional[str] = dataclasses.field(default=NA)
    strict: bool = dataclasses.field(default=NA)
    exact: bool = dataclasses.field(default=NA)
    cache: bool = dataclasses.field(default=NA)
    ambiguous: str = dataclasses.field(default=NA)

    @classmethod
    def from_dict(cls, dct: T_KWARGS):
        req_kwargs, opt_kwargs = cls._split_req_opt(dct)
        req_kwargs["expr"] = parse_expr(req_kwargs["expr"])
        return cls(**req_kwargs, **rm_na(**opt_kwargs))

    def to_polars(self) -> pl.Expr:
        return ensure_string(self.expr).to_datetime(
            **rm_na(
                format=self.format,
                time_unit=self.time_unit,
                time_zone=self.time_zone,
                strict=self.strict,
                exact=self.exact,
                cache=self.cache,
                ambiguous=self.ambiguous,
            )
        )


expr_enum_to_klass_mapping[ExprEnum.str_to_datetime.value] = StrToDatetime


@dataclasses.dataclass
class StrToDate(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.str.to_date.html
    """

    type: str = dataclasses.field(default=ExprEnum.str_to_date.value)
    expr: "T_EXPR" = dataclasses.field(default=REQ)
    format: T.Optional[str] = dataclasses.field(default=NA)
    strict: bool = dataclasses.field(default=NA)
    exact: bool = dataclasses.field(default=NA)
    cache: bool = dataclasses.field(default=NA)

    @classmethod
    def from_dict(cls, dct: T_KWARGS):
        req_kwargs, opt_kwargs = cls._split_req_opt(dct)
        req_kwargs["expr"] = parse_expr(req_kwargs["expr"])
        return cls(**req_kwargs, **rm_na(**opt_kwargs))

    def to_polars(self) -> pl.Expr:
        return ensure_string(self.expr).to_date(
            **rm_na(
                format=self.format,
                strict=self.strict,
                exact=self.exact,
                cache=self.cache,
            )
        )


expr_enum_to_klass_mapping[ExprEnum.str_to_date.value] = StrToDate


@dataclasses.dataclass
class StrZfill(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.str.zfill.html
    """

    type: str = dataclasses.field(default=ExprEnum.str_zfill.value)
    expr: "T_EXPR" = dataclasses.field(default=REQ)
    length: int = dataclasses.field(default=REQ)

    @classmethod
    def from_dict(cls, dct: T_KWARGS):
        return cls(
            expr=parse_expr(dct["expr"]),
            length=dct["length"],
        )

    def to_polars(self) -> pl.Expr:
        return ensure_string(self.expr).zfill(length=self.length)


expr_enum_to_klass_mapping[ExprEnum.str_zfill.value] = StrZfill


@dataclasses.dataclass
class StrPadStart(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.str.pad_start.html
    """

    type: str = dataclasses.field(default=ExprEnum.str_pad_start.value)
    expr: "T_EXPR" = dataclasses.field(default=REQ)
    length: int = dataclasses.field(default=REQ)
    fill_char: str = dataclasses.field(default=NA)

    @classmethod
    def from_dict(cls, dct: T_KWARGS):
        req_kwargs, opt_kwargs = cls._split_req_opt(dct)
        req_kwargs["expr"] = parse_expr(req_kwargs["expr"])
        return cls(**req_kwargs, **rm_na(**opt_kwargs))

    def to_polars(self) -> pl.Expr:
        return ensure_string(self.expr).pad_start(
            length=self.length,
            **rm_na(
                fill_char=self.fill_char,
            ),
        )


expr_enum_to_klass_mapping[ExprEnum.str_pad_start.value] = StrPadStart


@dataclasses.dataclass
class StrPadEnd(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.str.pad_end.html
    """

    type: str = dataclasses.field(default=ExprEnum.str_pad_end.value)
    expr: "T_EXPR" = dataclasses.field(default=REQ)
    length: int = dataclasses.field(default=REQ)
    fill_char: str = dataclasses.field(default=NA)

    @classmethod
    def from_dict(cls, dct: T_KWARGS):
        req_kwargs, opt_kwargs = cls._split_req_opt(dct)
        req_kwargs["expr"] = parse_expr(req_kwargs["expr"])
        return cls(**req_kwargs, **rm_na(**opt_kwargs))

    def to_polars(self) -> pl.Expr:
        return ensure_string(self.expr).pad_end(
            length=self.length,
            **rm_na(
                fill_char=self.fill_char,
            ),
        )


expr_enum_to_klass_mapping[ExprEnum.str_pad_end.value] = StrPadEnd


@dataclasses.dataclass
class StrToLowerCase(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.str.to_lowercase.html
    """

    type: str = dataclasses.field(default=ExprEnum.str_to_lowercase.value)
    expr: "T_EXPR" = dataclasses.field(default=REQ)

    @classmethod
    def from_dict(cls, dct: T_KWARGS):
        return cls(
            expr=parse_expr(dct["expr"]),
        )

    def to_polars(self) -> pl.Expr:
        return ensure_string(self.expr).to_lowercase()


expr_enum_to_klass_mapping[ExprEnum.str_to_lowercase.value] = StrToLowerCase


@dataclasses.dataclass
class StrToUpperCase(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.str.to_uppercase.html
    """

    type: str = dataclasses.field(default=ExprEnum.str_to_uppercase.value)
    expr: "T_EXPR" = dataclasses.field(default=REQ)

    @classmethod
    def from_dict(cls, dct: T_KWARGS):
        return cls(
            expr=parse_expr(dct["expr"]),
        )

    def to_polars(self) -> pl.Expr:
        return ensure_string(self.expr).to_uppercase()


expr_enum_to_klass_mapping[ExprEnum.str_to_uppercase.value] = StrToUpperCase


@dataclasses.dataclass
class StrToTitleCase(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.str.to_titlecase.html
    """

    type: str = dataclasses.field(default=ExprEnum.str_to_titlecase.value)
    expr: "T_EXPR" = dataclasses.field(default=REQ)

    @classmethod
    def from_dict(cls, dct: T_KWARGS):
        return cls(
            expr=parse_expr(dct["expr"]),
        )

    def to_polars(self) -> pl.Expr:
        return ensure_string(self.expr).to_titlecase()


expr_enum_to_klass_mapping[ExprEnum.str_to_titlecase.value] = StrToTitleCase


@dataclasses.dataclass
class StrHead(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.str.head.html
    """

    type: str = dataclasses.field(default=ExprEnum.str_head.value)
    expr: "T_EXPR" = dataclasses.field(default=REQ)
    n: int = dataclasses.field(default=REQ)

    @classmethod
    def from_dict(cls, dct: T_KWARGS):
        return cls(
            expr=parse_expr(dct["expr"]),
            n=dct["n"],
        )

    def to_polars(self) -> pl.Expr:
        return ensure_string(self.expr).head(n=self.n)


expr_enum_to_klass_mapping[ExprEnum.str_head.value] = StrHead


@dataclasses.dataclass
class StrTail(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.str.tail.html
    """

    type: str = dataclasses.field(default=ExprEnum.str_tail.value)
    expr: "T_EXPR" = dataclasses.field(default=REQ)
    n: int = dataclasses.field(default=REQ)

    @classmethod
    def from_dict(cls, dct: T_KWARGS):
        return cls(
            expr=parse_expr(dct["expr"]),
            n=dct["n"],
        )

    def to_polars(self) -> pl.Expr:
        return ensure_string(self.expr).tail(n=self.n)


expr_enum_to_klass_mapping[ExprEnum.str_tail.value] = StrTail


@dataclasses.dataclass
class StrSlice(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.str.slice.html
    """

    type: str = dataclasses.field(default=ExprEnum.str_slice.value)
    expr: "T_EXPR" = dataclasses.field(default=REQ)
    offset: int = dataclasses.field(default=REQ)
    length: int = dataclasses.field(default=REQ)

    @classmethod
    def from_dict(cls, dct: T_KWARGS):
        return cls(
            expr=parse_expr(dct["expr"]),
            offset=dct["offset"],
            length=dct["length"],
        )

    def to_polars(self) -> pl.Expr:
        return ensure_string(self.expr).slice(
            offset=self.offset,
            length=self.length,
        )


expr_enum_to_klass_mapping[ExprEnum.str_slice.value] = StrSlice


@dataclasses.dataclass
class StrReplace(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.str.replace.html
    """

    type: str = dataclasses.field(default=ExprEnum.str_replace.value)
    expr: "T_EXPR" = dataclasses.field(default=REQ)
    pattern: T.Union[str, "T_EXPR"] = dataclasses.field(default=REQ)
    value: T.Union[str, "T_EXPR"] = dataclasses.field(default=REQ)
    literal: bool = dataclasses.field(default=NA)
    n: int = dataclasses.field(default=NA)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        req_kwargs, opt_kwargs = cls._split_req_opt(dct)
        req_kwargs["expr"] = parse_expr(req_kwargs["expr"])
        req_kwargs["pattern"] = to_jsonpolars_other_expr(req_kwargs["pattern"])
        req_kwargs["value"] = to_jsonpolars_other_expr(req_kwargs["value"])
        return cls(**req_kwargs, **rm_na(**opt_kwargs))

    def to_polars(self) -> pl.Expr:
        return ensure_string(self.expr).replace(
            pattern=to_polars_other_expr(self.pattern),
            value=to_polars_other_expr(self.value),
            **rm_na(
                literal=self.literal,
                n=self.n,
            ),
        )


expr_enum_to_klass_mapping[ExprEnum.str_replace.value] = StrReplace


@dataclasses.dataclass
class StrReplaceAll(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.str.replace_all.html
    """

    type: str = dataclasses.field(default=ExprEnum.str_replace_all.value)
    expr: "T_EXPR" = dataclasses.field(default=REQ)
    pattern: T.Union[str, "T_EXPR"] = dataclasses.field(default=REQ)
    value: T.Union[str, "T_EXPR"] = dataclasses.field(default=REQ)
    literal: bool = dataclasses.field(default=NA)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        req_kwargs, opt_kwargs = cls._split_req_opt(dct)
        req_kwargs["expr"] = parse_expr(req_kwargs["expr"])
        req_kwargs["pattern"] = to_jsonpolars_other_expr(req_kwargs["pattern"])
        req_kwargs["value"] = to_jsonpolars_other_expr(req_kwargs["value"])
        return cls(**req_kwargs, **rm_na(**opt_kwargs))

    def to_polars(self) -> pl.Expr:
        return ensure_string(self.expr).replace_all(
            pattern=to_polars_other_expr(self.pattern),
            value=to_polars_other_expr(self.value),
            **rm_na(
                literal=self.literal,
            ),
        )


expr_enum_to_klass_mapping[ExprEnum.str_replace_all.value] = StrReplaceAll



# -*- coding: utf-8 -*-

import typing as T
import dataclasses

import polars as pl

from ..arg import REQ, NA, rm_na, T_KWARGS
from ..base_expr import ExprEnum, BaseExpr, expr_enum_to_klass_mapping, parse_expr

if T.TYPE_CHECKING:  # pragma: no cover
    from .api import T_EXPR


@dataclasses.dataclass
class Column(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/col.html
    """

    type: str = dataclasses.field(default=ExprEnum.column.value)
    name: str = dataclasses.field(default=REQ)

    def to_polars(self) -> pl.Expr:
        return pl.col(self.name)


expr_enum_to_klass_mapping[ExprEnum.column.value] = Column


@dataclasses.dataclass
class Alias(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.alias.html
    """

    type: str = dataclasses.field(default=ExprEnum.alias.value)
    name: str = dataclasses.field(default=REQ)
    expr: "T_EXPR" = dataclasses.field(default=REQ)

    @classmethod
    def from_dict(cls, dct: T_KWARGS):
        return cls(
            name=dct["name"],
            expr=parse_expr(dct["expr"]),
        )

    def to_polars(self) -> pl.Expr:
        return self.expr.to_polars().alias(self.name)


expr_enum_to_klass_mapping[ExprEnum.alias.value] = Alias



# -*- coding: utf-8 -*-

import typing as T
from ..arg import _PRE
from .manipulation import Cast
from .list import List
from .list import ListGet
from .list import ListEval
from .dt import Datetime
from .dt import DtToString
from .dt import DtYear
from .dt import DtQuarter
from .dt import DtMonth
from .dt import DtDay
from .dt import DtHour
from .dt import DtMinute
from .dt import DtSecond
from .dt import DtNanoSecond
from .dt import DtEpoch
from .dt import DtTimestamp
from .dt import DtTotalDays
from .dt import DtTotalHours
from .dt import DtTotalMinutes
from .dt import DtTotalSeconds
from .dt import DtTotalMilliSeconds
from .dt import DtTotalMicroSeconds
from .dt import DtTotalNanoSeconds
from .dt import DtTruncate
from .operator import Plus
from .operator import Minus
from .operator import Multiply
from .operator import TrueDiv
from .operator import FloorDiv
from .operator import Negative
from .operator import Pow
from .operator import Equal
from .operator import NotEqual
from .operator import GreatThan
from .operator import GreatThanOrEqual
from .operator import LessThan
from .operator import LessThanOrEqual
from .operator import LogicalAnd
from .operator import LogicalOr
from .string import String
from .string import Split
from .string import StrJoin
from .string import StrContains
from .string import StrDecode
from .string import StrEncode
from .string import StrStartsWith
from .string import StrEndsWith
from .string import StrToDatetime
from .string import StrToDate
from .string import StrZfill
from .string import StrPadStart
from .string import StrPadEnd
from .string import StrToLowerCase
from .string import StrToUpperCase
from .string import StrToTitleCase
from .string import StrHead
from .string import StrTail
from .string import StrSlice
from .string import StrReplace
from .string import StrReplaceAll
from .function import Lit
from .function import ConcatStr
from .function import ConcatList
from .function import FuncStruct
from .function import Format
from .function import FuncDate
from .function import FuncDatetime
from .function import Element
from .column import Column
from .column import Alias
from .struct import Struct
from .struct import StructField
from .struct import StructRenameFields
from .struct import StructWithFields

T_EXPR = T.Union[
    _PRE,
    Cast,
    List,
    ListGet,
    ListEval,
    Datetime,
    DtToString,
    DtYear,
    DtQuarter,
    DtMonth,
    DtDay,
    DtHour,
    DtMinute,
    DtSecond,
    DtNanoSecond,
    DtEpoch,
    DtTimestamp,
    DtTotalDays,
    DtTotalHours,
    DtTotalMinutes,
    DtTotalSeconds,
    DtTotalMilliSeconds,
    DtTotalMicroSeconds,
    DtTotalNanoSeconds,
    DtTruncate,
    Plus,
    Minus,
    Multiply,
    TrueDiv,
    FloorDiv,
    Negative,
    Pow,
    Equal,
    NotEqual,
    GreatThan,
    GreatThanOrEqual,
    LessThan,
    LessThanOrEqual,
    LogicalAnd,
    LogicalOr,
    String,
    Split,
    StrJoin,
    StrContains,
    StrDecode,
    StrEncode,
    StrStartsWith,
    StrEndsWith,
    StrToDatetime,
    StrToDate,
    StrZfill,
    StrPadStart,
    StrPadEnd,
    StrToLowerCase,
    StrToUpperCase,
    StrToTitleCase,
    StrHead,
    StrTail,
    StrSlice,
    StrReplace,
    StrReplaceAll,
    Lit,
    ConcatStr,
    ConcatList,
    FuncStruct,
    Format,
    FuncDate,
    FuncDatetime,
    Element,
    Column,
    Alias,
    Struct,
    StructField,
    StructRenameFields,
    StructWithFields,
]



# -*- coding: utf-8 -*-

from datetime import datetime, date, timezone

from jsonpolars.expr import api as expr
from jsonpolars.chain import chain, PRE
from jsonpolars.tests.expr_case import Case


case_split_1 = Case(
    input_records=[
        {"id": "a-1"},
    ],
    expr=expr.Split(
        expr=expr.Column(name="id"),
        by="-",
    ),
    expected_output_records=[
        {"id": ["a", "1"]},
    ],
)
case_split_2 = Case(
    input_records=[
        {"id": "a-1"},
    ],
    expr=expr.Split(
        expr=expr.String(expr=expr.Column(name="id")),
        by="-",
    ),
    expected_output_records=[
        {"id": ["a", "1"]},
    ],
)
case_split_3 = Case(
    input_records=[
        {"id": "a-1"},
    ],
    expr=chain(
        expr.Column(name="id"),
        expr.Split(expr=PRE, by="-"),
    ),
    expected_output_records=[
        {"id": ["a", "1"]},
    ],
)
case_join = Case(
    input_records=[
        {"foo": "1"},
        {"foo": "2"},
        {"foo": "3"},
    ],
    expr=expr.StrJoin(
        delimiter="-",
        expr=expr.Column(name="foo"),
    ),
    expected_output_records=[
        {"foo": "1-2-3"},
        {"foo": "1-2-3"},
        {"foo": "1-2-3"},
    ],
)
case_contains_1 = Case(
    input_records=[
        {"foo": "123-456-7890"},
    ],
    expr=expr.StrContains(
        expr=expr.Column(name="foo"),
        pattern="456",
    ),
    expected_output_records=[
        {"foo": True},
    ],
)
case_contains_2 = Case(
    input_records=[
        {"foo": "123-456-7890"},
    ],
    expr=expr.StrContains(
        expr=expr.Column(name="foo"),
        pattern=r"\d{4}",
    ),
    expected_output_records=[
        {"foo": True},
    ],
)
case_contains_3 = Case(
    input_records=[
        {"foo": "123-456-7890"},
    ],
    expr=expr.StrContains(
        expr=expr.Column(name="foo"),
        pattern=expr.Lit(value=r"\d{10}"),
    ),
    expected_output_records=[
        {"foo": False},
    ],
)
case_decode = Case(
    input_records=[
        {"color": "ffff00"},
    ],
    expr=expr.StrDecode(
        expr=expr.Column(name="color"),
        encoding="hex",
    ),
    expected_output_records=[
        {"color": b"\xff\xff\x00"},
    ],
)
case_encode = Case(
    input_records=[
        {"word": "foo"},
    ],
    expr=expr.StrEncode(
        expr=expr.Column(name="word"),
        encoding="hex",
    ),
    expected_output_records=[
        {"word": "666f6f"},
    ],
)
case_starts_with_1 = Case(
    input_records=[
        {"foo": "Alice"},
    ],
    expr=expr.StrStartsWith(
        expr=expr.Column(name="foo"),
        prefix="A",
    ),
    expected_output_records=[
        {"foo": True},
    ],
)
case_starts_with_2 = Case(
    input_records=[
        {"foo": "Alice"},
    ],
    expr=expr.StrStartsWith(
        expr=expr.Column(name="foo"),
        prefix=expr.Lit(value="B"),
    ),
    expected_output_records=[
        {"foo": False},
    ],
)
case_starts_with_3 = Case(
    input_records=[
        {"foo": "Alice", "prefix": "A"},
    ],
    expr=expr.StrStartsWith(
        expr=expr.Column(name="foo"),
        prefix=expr.Column(name="prefix"),
    ),
    expected_output_records=[
        {"foo": True, "prefix": "A"},
    ],
)
case_ends_with_1 = Case(
    input_records=[
        {"foo": "Alice"},
    ],
    expr=expr.StrEndsWith(
        expr=expr.Column(name="foo"),
        suffix="e",
    ),
    expected_output_records=[
        {"foo": True},
    ],
)
case_ends_with_2 = Case(
    input_records=[
        {"foo": "Alice"},
    ],
    expr=expr.StrEndsWith(
        expr=expr.Column(name="foo"),
        suffix=expr.Lit(value="f"),
    ),
    expected_output_records=[
        {"foo": False},
    ],
)
case_ends_with_3 = Case(
    input_records=[
        {"foo": "Alice", "suffix": "e"},
    ],
    expr=expr.StrEndsWith(
        expr=expr.Column(name="foo"),
        suffix=expr.Column(name="suffix"),
    ),
    expected_output_records=[
        {"foo": True, "suffix": "e"},
    ],
)
case_to_datetime_1 = Case(
    input_records=[
        {"time": "2024-08-15T10:45:28"},
    ],
    expr=expr.StrToDatetime(
        expr=expr.Column(name="time"),
    ),
    expected_output_records=[
        {"time": datetime(2024, 8, 15, 10, 45, 28)},
    ],
)
# unit test doesn't pass in lower version of polars, however the code works fine
# case_to_datetime_2 = Case(
#     input_records=[
#         {"time": "2024-08-15T10:45:28"},
#     ],
#     expr=expr.StrToDatetime(
#         expr=expr.Column(name="time"),
#         time_zone="UTC",
#     ),
#     expected_output_records=[
#         {"time": datetime(2024, 8, 15, 10, 45, 28, tzinfo=timezone.utc)},
#     ],
# )
case_to_date = Case(
    input_records=[
        {"date": "2024-08-15"},
    ],
    expr=expr.StrToDate(
        expr=expr.Column(name="date"),
    ),
    expected_output_records=[
        {"date": date(2024, 8, 15)},
    ],
)
case_zfill = Case(
    input_records=[
        {"id": "1"},
    ],
    expr=expr.StrZfill(
        expr=expr.Column(name="id"),
        length=3,
    ),
    expected_output_records=[
        {"id": "001"},
    ],
)
case_pad_start = Case(
    input_records=[
        {"id": "1"},
    ],
    expr=expr.StrPadStart(
        expr=expr.Column(name="id"),
        length=3,
        fill_char="0",
    ),
    expected_output_records=[
        {"id": "001"},
    ],
)
case_pad_end = Case(
    input_records=[
        {"id": "1"},
    ],
    expr=expr.StrPadEnd(
        expr=expr.Column(name="id"),
        length=3,
        fill_char="0",
    ),
    expected_output_records=[
        {"id": "100"},
    ],
)
case_to_lowercase = Case(
    input_records=[
        {"id": "Env"},
    ],
    expr=expr.StrToLowerCase(
        expr=expr.Column(name="id"),
    ),
    expected_output_records=[
        {"id": "env"},
    ],
)
case_to_uppercase = Case(
    input_records=[
        {"id": "Env"},
    ],
    expr=expr.StrToUpperCase(
        expr=expr.Column(name="id"),
    ),
    expected_output_records=[
        {"id": "ENV"},
    ],
)
case_to_titlecase = Case(
    input_records=[
        {"id": "env"},
    ],
    expr=expr.StrToTitleCase(
        expr=expr.Column(name="id"),
    ),
    expected_output_records=[
        {"id": "Env"},
    ],
)
case_head = Case(
    input_records=[
        {"id": "Env"},
    ],
    expr=expr.StrHead(
        expr=expr.Column(name="id"),
        n=1,
    ),
    expected_output_records=[
        {"id": "E"},
    ],
)
case_tail = Case(
    input_records=[
        {"id": "Env"},
    ],
    expr=expr.StrTail(
        expr=expr.Column(name="id"),
        n=1,
    ),
    expected_output_records=[
        {"id": "v"},
    ],
)
case_slice = Case(
    input_records=[
        {"time": "2024-08-15 10:45:28"},
    ],
    expr=expr.StrSlice(
        expr=expr.Column(name="time"),
        offset=5,
        length=5,
    ),
    expected_output_records=[
        {"time": "08-15"},
    ],
)
case_replace_1 = Case(
    input_records=[
        {"text": "current env is prod, target env is prod"},
    ],
    expr=expr.StrReplace(
        expr=expr.Column(name="text"),
        pattern="prod",
        value="dev",
        literal=True,
    ),
    expected_output_records=[
        {"text": "current env is dev, target env is prod"},
    ],
)
case_replace_2 = Case(
    input_records=[
        {"text": "current env is prod, target env is prod"},
    ],
    expr=expr.StrReplace(
        expr=expr.Column(name="text"),
        pattern=expr.Lit(value="prod"),
        value=expr.Lit(value="dev"),
        literal=True,
    ),
    expected_output_records=[
        {"text": "current env is dev, target env is prod"},
    ],
)
case_replace_all_1 = Case(
    input_records=[
        {"text": "current env is prod, target env is prod"},
    ],
    expr=expr.StrReplaceAll(
        expr=expr.Column(name="text"),
        pattern=expr.Lit(value="prod"),
        value=expr.Lit(value="dev"),
        literal=True,
    ),
    expected_output_records=[
        {"text": "current env is dev, target env is dev"},
    ],
)
case_replace_all_2 = Case(
    input_records=[
        {"text": "current env is prod, target env is prod"},
    ],
    expr=chain(
        expr.Column(name="text"),
        expr.StrReplaceAll(
            expr=PRE,
            pattern=expr.Lit(value="prod"),
            value=expr.Lit(value="dev"),
            literal=True,
        ),
    ),
    expected_output_records=[
        {"text": "current env is dev, target env is dev"},
    ],
)


def test():
    print("")

    case_split_1.run_with_columns_test()
    case_split_2.run_with_columns_test()
    case_split_3.run_with_columns_test()
    case_join.run_with_columns_test()
    case_contains_1.run_with_columns_test()
    case_contains_2.run_with_columns_test()
    case_contains_3.run_with_columns_test()
    case_decode.run_with_columns_test()
    case_encode.run_with_columns_test()
    case_starts_with_1.run_with_columns_test()
    case_starts_with_2.run_with_columns_test()
    case_starts_with_3.run_with_columns_test()
    case_ends_with_1.run_with_columns_test()
    case_ends_with_2.run_with_columns_test()
    case_ends_with_3.run_with_columns_test()
    case_to_datetime_1.run_with_columns_test()
    # case_to_datetime_2.run_with_columns_test()
    case_to_date.run_with_columns_test()
    case_zfill.run_with_columns_test()
    case_pad_start.run_with_columns_test()
    case_pad_end.run_with_columns_test()
    case_to_lowercase.run_with_columns_test()
    case_to_uppercase.run_with_columns_test()
    case_to_titlecase.run_with_columns_test()
    case_head.run_with_columns_test()
    case_tail.run_with_columns_test()
    case_slice.run_with_columns_test()
    case_replace_1.run_with_columns_test()
    case_replace_2.run_with_columns_test()
    case_replace_all_1.run_with_columns_test()
    case_replace_all_2.run_with_columns_test()


if __name__ == "__main__":
    from jsonpolars.tests import run_cov_test

    run_cov_test(__file__, "jsonpolars.expr.string", preview=False)



# -*- coding: utf-8 -*-

"""
这个模块中有一系列的用于 serde 的函数.

- 凡是以 ``to_jsonpolars`` 开头的函数都是用于将不同的 Python 数据结构转化成类 BaseExpr 对象.
- 凡是以 ``to_polars`` 开头的函数都是用于将不同的 Python Literal 和 BaseExpr 对象转化为 polars Expr 对象.
"""

import typing as T

import polars as pl

from .base_expr import BaseExpr, parse_expr
from .typehint import (
    PythonLiteral,
    IntoExpr,
    OtherExpr,
)


if T.TYPE_CHECKING:  # pragma: no cover
    from .expr.api import T_EXPR


def to_jsonpolars_into_expr(
    expr_like: T.Union[str, dict, "T_EXPR"],
) -> "IntoExpr":
    if isinstance(expr_like, str):
        return expr_like
    elif isinstance(expr_like, dict):
        return parse_expr(expr_like)
    elif isinstance(expr_like, BaseExpr):
        return expr_like
    else:  # pragma: no cover
        raise NotImplementedError(f"Unsupported type: {type(expr_like)}")


def to_polars_into_expr(
    expr_like: T.Union[str, "T_EXPR"],
) -> T.Union[str, "pl.Expr"]:
    if isinstance(expr_like, BaseExpr):
        return expr_like.to_polars()
    else:  # pragma: no cover
        return expr_like


def batch_to_jsonpolars_into_exprs(
    exprs: T.Iterable[T.Union[str, dict, "T_EXPR"]],
) -> T.List["IntoExpr"]:
    """
    .. note::

        Intentionally not using list comprehension here. So that it tells you
        which expression is causing the error.
    """
    new_exprs = list()
    for ex in exprs:
        new_exprs.append(to_jsonpolars_into_expr(ex))
    return new_exprs


def batch_to_jsonpolars_named_into_exprs(
    named_exprs: T.Dict[str, T.Union[str, dict, "T_EXPR"]],
) -> T.Dict[str, "IntoExpr"]:
    """
    .. note::

        Intentionally not using list comprehension here. So that it tells you
        which expression is causing the error.
    """
    new_named_exprs = dict()
    for name, ex in named_exprs.items():
        new_named_exprs[name] = to_jsonpolars_into_expr(ex)
    return new_named_exprs


def batch_to_polars_into_exprs(
    exprs: T.Iterable[T.Union[str, "T_EXPR"]],
) -> T.List[T.Union[str, "pl.Expr"]]:
    """
    .. note::

        Intentionally not using list comprehension here. So that it tells you
        which expression is causing the error.
    """
    new_exprs = list()
    for ex in exprs:
        new_exprs.append(to_polars_into_expr(ex))
    return new_exprs


def batch_to_polars_named_into_exprs(
    named_exprs: T.Dict[str, T.Union[str, "T_EXPR"]],
) -> T.Dict[str, T.Union[str, "pl.Expr"]]:
    """
    .. note::

        Intentionally not using list comprehension here. So that it tells you
        which expression is causing the error.
    """
    new_named_exprs = dict()
    for name, ex in named_exprs.items():
        new_named_exprs[name] = to_polars_into_expr(ex)
    return new_named_exprs


def to_jsonpolars_other_expr(
    expr_like: T.Union["PythonLiteral", dict, "T_EXPR"]
) -> "OtherExpr":
    if isinstance(expr_like, dict):
        return parse_expr(expr_like)
    else:
        return expr_like


def to_polars_other_expr(
    expr_like: "OtherExpr",
) -> T.Union["PythonLiteral", "pl.Expr"]:
    if isinstance(expr_like, BaseExpr):
        return expr_like.to_polars()
    else:
        return expr_like


class PolarsTypeNameEnum:
    Int8 = "Int8"
    Int16 = "Int16"
    Int32 = "Int32"
    Int64 = "Int64"
    Float32 = "Float32"
    Float64 = "Float64"
    Decimal = "Decimal"
    String = "String"
    Binary = "Binary"
    Boolean = "Boolean"
    Null = "Null"
    Datetime = "Datetime"
    Date = "Date"
    Duration = "Duration"
    List = "List"
    Struct = "Struct"


str_to_polars_dtype_mapping = {
    PolarsTypeNameEnum.Int8: pl.Int8,
    PolarsTypeNameEnum.Int16: pl.Int16,
    PolarsTypeNameEnum.Int32: pl.Int32,
    PolarsTypeNameEnum.Int64: pl.Int64,
    PolarsTypeNameEnum.Float32: pl.Float32,
    PolarsTypeNameEnum.Float64: pl.Float64,
    PolarsTypeNameEnum.Decimal: pl.Decimal,
    PolarsTypeNameEnum.String: pl.String,
    PolarsTypeNameEnum.Binary: pl.Binary,
    PolarsTypeNameEnum.Boolean: pl.Boolean,
    PolarsTypeNameEnum.Null: pl.Null,
    PolarsTypeNameEnum.Datetime: pl.Datetime,
    PolarsTypeNameEnum.Date: pl.Date,
    PolarsTypeNameEnum.Duration: pl.Duration,
}

polars_dtype_to_str_mapping = {
    v: k
    for k, v in str_to_polars_dtype_mapping.items()
}



# -*- coding: utf-8 -*-

import typing as T
import dataclasses

import polars as pl
from simpletype.api import json_type_to_simple_type

from ..arg import REQ, NA, rm_na, T_KWARGS
from ..base_expr import ExprEnum, BaseExpr, expr_enum_to_klass_mapping, parse_expr
from ..utils_expr import (
    batch_to_jsonpolars_into_exprs,
    batch_to_jsonpolars_named_into_exprs,
    batch_to_polars_into_exprs,
    batch_to_polars_named_into_exprs,
    to_jsonpolars_other_expr,
    str_to_polars_dtype_mapping,
    polars_dtype_to_str_mapping,
)

if T.TYPE_CHECKING:  # pragma: no cover
    from .api import T_EXPR
    from ..typehint import IntoExpr, DatetimeElementExpr


@dataclasses.dataclass
class Lit(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.lit.html
    """

    type: str = dataclasses.field(default=ExprEnum.func_lit.value)
    value: T.Any = dataclasses.field(default=REQ)
    dtype: T.Union[str, "pl.DataType", T.Type["pl.DataType"]] = dataclasses.field(
        default=NA
    )
    allow_object: bool = dataclasses.field(default=NA)

    def to_dict(self) -> T_KWARGS:
        dct = super().to_dict()
        if "dtype" in dct:
            if isinstance(self.dtype, str):
                dtype = self.dtype
            else:
                try:
                    dtype = polars_dtype_to_str_mapping[type(self.dtype)]
                except KeyError:
                    dtype = polars_dtype_to_str_mapping[self.dtype]
            dct["dtype"] = dtype
        return dct

    def to_polars(self) -> pl.Expr:
        if isinstance(self.dtype, str):
            dtype = str_to_polars_dtype_mapping[self.dtype]
        else:
            dtype = self.dtype
        return pl.lit(
            value=self.value,
            **rm_na(
                dtype=dtype,
                allow_object=self.allow_object,
            ),
        )


expr_enum_to_klass_mapping[ExprEnum.func_lit.value] = Lit


@dataclasses.dataclass
class ConcatStr(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.concat_str.html
    """

    type: str = dataclasses.field(default=ExprEnum.func_concat_str.value)
    exprs: T.List["IntoExpr"] = dataclasses.field(default_factory=list)
    separator: str = dataclasses.field(default=NA)
    ignore_nulls: bool = dataclasses.field(default=NA)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        req_kwargs, opt_kwargs = cls._split_req_opt(dct)
        opt_kwargs["exprs"] = batch_to_jsonpolars_into_exprs(
            opt_kwargs.get("exprs", [])
        )
        return cls(**req_kwargs, **rm_na(**opt_kwargs))

    def to_polars(self) -> pl.Expr:
        return pl.concat_str(
            exprs=batch_to_polars_into_exprs(self.exprs),
            **rm_na(
                separator=self.separator,
                ignore_nulls=self.ignore_nulls,
            ),
        )


expr_enum_to_klass_mapping[ExprEnum.func_concat_str.value] = ConcatStr


@dataclasses.dataclass
class ConcatList(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.concat_list.html
    """

    type: str = dataclasses.field(default=ExprEnum.func_concat_list.value)
    exprs: T.List["IntoExpr"] = dataclasses.field(default_factory=list)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        return cls(
            exprs=batch_to_jsonpolars_into_exprs(dct.get("exprs", list())),
        )

    def to_polars(self) -> pl.Expr:
        return pl.concat_list(
            exprs=batch_to_polars_into_exprs(self.exprs),
        )


expr_enum_to_klass_mapping[ExprEnum.func_concat_list.value] = ConcatList


@dataclasses.dataclass
class FuncStruct(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.struct.html
    """

    type: str = dataclasses.field(default=ExprEnum.func_struct.value)
    exprs: T.List["IntoExpr"] = dataclasses.field(default_factory=list)
    named_exprs: T.Dict[str, "IntoExpr"] = dataclasses.field(default_factory=dict)
    schema: T.Optional[T.Dict[str, T.Dict[str, T.Any]]] = dataclasses.field(default=NA)
    eager: bool = dataclasses.field(default=NA)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        req_kwargs, opt_kwargs = cls._split_req_opt(dct)
        opt_kwargs["exprs"] = batch_to_jsonpolars_into_exprs(
            opt_kwargs.get("exprs", [])
        )
        opt_kwargs["named_exprs"] = batch_to_jsonpolars_named_into_exprs(
            opt_kwargs.get("named_exprs", {})
        )
        return cls(**req_kwargs, **rm_na(**opt_kwargs))

    def to_polars(self) -> pl.Expr:
        kwargs = {}
        if isinstance(self.schema, dict):
            kwargs["schema"] = {
                k: json_type_to_simple_type(v).to_polars()
                for k, v in self.schema.items()
            }
        return pl.struct(
            *batch_to_polars_into_exprs(self.exprs),
            **batch_to_polars_named_into_exprs(self.named_exprs),
            **rm_na(eager=self.eager, **kwargs),
        )


expr_enum_to_klass_mapping[ExprEnum.func_struct.value] = FuncStruct


@dataclasses.dataclass
class Format(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.format.html
    """

    type: str = dataclasses.field(default=ExprEnum.func_format.value)
    f_string: str = dataclasses.field(default=REQ)
    exprs: T.List["IntoExpr"] = dataclasses.field(default_factory=list)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        req_kwargs, opt_kwargs = cls._split_req_opt(dct)
        opt_kwargs["exprs"] = batch_to_jsonpolars_into_exprs(
            opt_kwargs.get("exprs", [])
        )
        return cls(**req_kwargs, **rm_na(**opt_kwargs))

    def to_polars(self) -> pl.Expr:
        return pl.format(
            self.f_string,
            *batch_to_polars_into_exprs(self.exprs),
        )


expr_enum_to_klass_mapping[ExprEnum.func_format.value] = Format


@dataclasses.dataclass
class FuncDate(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.date.html
    """

    type: str = dataclasses.field(default=ExprEnum.func_date.value)
    year: "DatetimeElementExpr" = dataclasses.field(default=REQ)
    month: "DatetimeElementExpr" = dataclasses.field(default=REQ)
    day: "DatetimeElementExpr" = dataclasses.field(default=REQ)

    def to_polars(self) -> pl.Expr:
        return pl.date(
            year=self.year,
            month=self.month,
            day=self.day,
        )


expr_enum_to_klass_mapping[ExprEnum.func_date.value] = FuncDate


@dataclasses.dataclass
class FuncDatetime(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.datetime.html
    """

    type: str = dataclasses.field(default=ExprEnum.func_datetime.value)
    year: "DatetimeElementExpr" = dataclasses.field(default=REQ)
    month: "DatetimeElementExpr" = dataclasses.field(default=REQ)
    day: "DatetimeElementExpr" = dataclasses.field(default=REQ)
    hour: "DatetimeElementExpr" = dataclasses.field(default=NA)
    minute: "DatetimeElementExpr" = dataclasses.field(default=NA)
    second: "DatetimeElementExpr" = dataclasses.field(default=NA)
    microsecond: "DatetimeElementExpr" = dataclasses.field(default=NA)
    time_unit: str = dataclasses.field(default=NA)
    time_zone: T.Optional[str] = dataclasses.field(default=NA)
    ambiguous: str = dataclasses.field(default=NA)

    def to_polars(self) -> pl.Expr:
        return pl.datetime(
            year=self.year,
            month=self.month,
            day=self.day,
            **rm_na(
                hour=self.hour,
                minute=self.minute,
                second=self.second,
                microsecond=self.microsecond,
                time_unit=self.time_unit,
                time_zone=self.time_zone,
                ambiguous=self.ambiguous,
            ),
        )


expr_enum_to_klass_mapping[ExprEnum.func_datetime.value] = FuncDatetime


@dataclasses.dataclass
class Element(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.element.html
    """

    type: str = dataclasses.field(default=ExprEnum.func_element.value)

    def to_polars(self) -> pl.Expr:
        return pl.element()


expr_enum_to_klass_mapping[ExprEnum.func_element.value] = Element



# -*- coding: utf-8 -*-

import polars as pl
from datetime import date, datetime

from simpletype.api import Integer, Struct
from jsonpolars.expr import api as expr
from jsonpolars.chain import chain, PRE
from jsonpolars.utils_expr import PolarsTypeNameEnum
from jsonpolars.tests.expr_case import Case


case_lit_1 = Case(
    input_records=[
        {"a": 1},
    ],
    expr=expr.Alias(
        expr=expr.Lit(
            value=111,
        ),
        name="a",
    ),
    expected_output_records=[
        {"a": 111},
    ],
)
case_lit_2 = Case(
    input_records=[
        {"a": 1},
    ],
    expr=expr.Alias(
        expr=expr.Lit(
            value=111,
            dtype=PolarsTypeNameEnum.Int8,
        ),
        name="a",
    ),
    expected_output_records=[
        {"a": 111},
    ],
)
case_lit_3 = Case(
    input_records=[
        {"a": 1},
    ],
    expr=expr.Alias(
        expr=expr.Lit(
            value=222,
            dtype=pl.Int32,
        ),
        name="a",
    ),
    expected_output_records=[
        {"a": 222},
    ],
)
case_lit_4 = Case(
    input_records=[
        {"a": 1},
    ],
    expr=expr.Alias(
        expr=expr.Lit(
            value=333,
            dtype=pl.Int64(),
        ),
        name="a",
    ),
    expected_output_records=[
        {"a": 333},
    ],
)
case_concat_str = Case(
    input_records=[
        {"col_1": "a", "col_2": "b"},
    ],
    expr=expr.Alias(
        expr=expr.ConcatStr(
            exprs=[
                "col_1",
                expr.Column(name="col_2"),
            ],
            separator="-",
        ),
        name="col_3",
    ),
    expected_output_records=[
        {"col_1": "a", "col_2": "b", "col_3": "a-b"},
    ],
)
case_concat_list = Case(
    input_records=[
        {"col_1": [1, 2], "col_2": [3, 4]},
    ],
    expr=expr.Alias(
        expr=expr.ConcatList(
            exprs=[
                "col_1",
                expr.Column(name="col_2"),
            ],
        ),
        name="col_3",
    ),
    expected_output_records=[
        {"col_1": [1, 2], "col_2": [3, 4], "col_3": [1, 2, 3, 4]},
    ],
)
case_func_struct_1 = Case(
    input_records=[
        {
            "a": 1,
            "b": 2,
            "c": 3,
        },
    ],
    expr=expr.Alias(
        expr=expr.FuncStruct(
            exprs=["a", "b"],
        ),
        name="data",
    ),
    expected_output_records=[
        {"a": 1, "b": 2, "c": 3, "data": {"a": 1, "b": 2}},
    ],
)
case_func_struct_2 = Case(
    input_records=[
        {
            "a": 1,
            "b": 2,
            "c": 3,
        },
    ],
    expr=expr.Alias(
        expr=expr.FuncStruct(
            exprs=[
                expr.Alias(expr=expr.Column(name="a"), name="x"),
                expr.Alias(expr=expr.Column(name="b"), name="y"),
            ],
        ),
        name="data",
    ),
    expected_output_records=[
        {"a": 1, "b": 2, "c": 3, "data": {"x": 1, "y": 2}},
    ],
)
case_func_struct_3 = Case(
    input_records=[
        {
            "a": 1,
            "b": 2,
            "c": 3,
        },
    ],
    expr=expr.Alias(
        expr=expr.FuncStruct(
            named_exprs=dict(
                x=expr.Column(name="a"),
                y=expr.Column(name="b"),
            ),
        ),
        name="data",
    ),
    expected_output_records=[
        {"a": 1, "b": 2, "c": 3, "data": {"x": 1, "y": 2}},
    ],
)
case_func_struct_4 = Case(
    input_records=[
        {"data": {"a": 1, "b": 2, "c": 3}},
    ],
    expr=expr.Alias(
        expr=expr.FuncStruct(
            named_exprs=dict(
                x=expr.StructField(expr=expr.Column(name="data"), name="a"),
                y=expr.StructField(expr=expr.Column(name="data"), name="b"),
            ),
        ),
        name="res",
    ),
    expected_output_records=[
        {"data": {"a": 1, "b": 2, "c": 3}, "res": {"x": 1, "y": 2}},
    ],
)
case_func_struct_5 = Case(
    input_records=[
        {"data": {"a": 1, "b": 2, "c": 3}},
    ],
    expr=expr.Alias(
        expr=expr.FuncStruct(
            exprs=[
                expr.StructField(
                    expr=expr.Column(name="data"),
                    name="a",
                    more_names=["b"],
                )
            ]
        ),
        name="res",
    ),
    expected_output_records=[
        {"data": {"a": 1, "b": 2, "c": 3}, "res": {"a": 1, "b": 2}},
    ],
)
case_func_struct_6 = Case(
    input_records=[
        {
            "a": 1,
            "b": 2,
            "c": 3,
        },
    ],
    expr=expr.Alias(
        expr=expr.FuncStruct(
            exprs=["a", "b"],
            schema={
                "a": Integer().to_dict(),
                "b": Integer().to_dict(),
            },
        ),
        name="data",
    ),
    expected_output_records=[
        {"a": 1, "b": 2, "c": 3, "data": {"a": 1, "b": 2}},
    ],
)
case_func_struct_7 = Case(
    input_records=[
        {
            "a": 1,
            "b": 2,
            "struct1": {
                "c": 3,
                "d": 4,
                "struct2": {
                    "e": 5,
                    "f": 6,
                },
            },
        },
    ],
    expr=expr.Alias(
        expr=expr.FuncStruct(
            named_exprs=dict(
                c=expr.StructField(
                    expr=expr.Column(name="struct1"),
                    name="c",
                ),
                struct2=expr.FuncStruct(
                    named_exprs=dict(
                        e=expr.StructField(
                            expr=expr.StructField(
                                expr=expr.Column(name="struct1"),
                                name="struct2",
                            ),
                            name="e",
                        ),
                    )
                ),
            )
        ),
        name="struct1",
    ),
    expected_output_records=[
        {
            "a": 1,
            "b": 2,
            "struct1": {
                "c": 3,
                "struct2": {
                    "e": 5,
                },
            },
        },
    ],
)
case_func_struct_8 = Case(
    input_records=[
        {
            "a": 1,
            "b": 2,
            "struct1": {
                "c": 3,
                "d": 4,
                "struct2": {
                    "e": 5,
                    "f": 6,
                },
            },
        },
    ],
    expr=chain(
        expr.FuncStruct(
            named_exprs=dict(
                c=expr.StructField(
                    expr=expr.Column(name="struct1"),
                    name="c",
                ),
                struct2=expr.FuncStruct(
                    named_exprs=dict(
                        e=chain(
                            expr.Column(name="struct1"),
                            expr.StructField(expr=PRE, name="struct2"),
                            expr.StructField(expr=PRE, name="e"),
                        )
                    ),
                ),
            ),
        ),
        expr.Alias(expr=PRE, name="struct1"),
    ),
    expected_output_records=[
        {
            "a": 1,
            "b": 2,
            "struct1": {
                "c": 3,
                "struct2": {
                    "e": 5,
                },
            },
        },
    ],
)
case_format = Case(
    input_records=[
        {"col_1": "a", "col_2": 1},
    ],
    expr=expr.Alias(
        expr=expr.Format(
            f_string="{}-{}",
            exprs=[
                "col_1",
                expr.Column(name="col_2"),
            ],
        ),
        name="col_3",
    ),
    expected_output_records=[
        {"col_1": "a", "col_2": 1, "col_3": "a-1"},
    ],
)
case_date_1 = Case(
    input_records=[
        {"year": 2015, "month": 1, "day": 15},
    ],
    expr=expr.Alias(
        expr=expr.FuncDate(
            year=2008,
            month=7,
            day=23,
        ),
        name="date",
    ),
    expected_output_records=[
        {"year": 2015, "month": 1, "day": 15, "date": date(2008, 7, 23)},
    ],
)
case_date_2 = Case(
    input_records=[
        {"year": 2015, "month": 1, "day": 15},
    ],
    expr=expr.Alias(
        expr=expr.FuncDate(
            year="year",
            month="month",
            day="day",
        ),
        name="date",
    ),
    expected_output_records=[
        {"year": 2015, "month": 1, "day": 15, "date": date(2015, 1, 15)},
    ],
)
case_datetime_2 = Case(
    input_records=[
        {"year": 2015, "month": 1, "day": 15},
    ],
    expr=expr.Alias(
        expr=expr.FuncDatetime(
            year=2008,
            month=7,
            day=23,
        ),
        name="date",
    ),
    expected_output_records=[
        {"year": 2015, "month": 1, "day": 15, "date": datetime(2008, 7, 23)},
    ],
)


def test():
    print("")

    case_lit_1.run_with_columns_test()
    case_lit_2.run_with_columns_test()
    case_lit_3.run_with_columns_test()
    case_lit_4.run_with_columns_test()
    case_concat_str.run_with_columns_test()
    case_concat_list.run_with_columns_test()
    case_func_struct_1.run_with_columns_test()
    case_func_struct_2.run_with_columns_test()
    case_func_struct_3.run_with_columns_test()
    case_func_struct_4.run_with_columns_test()
    case_func_struct_5.run_with_columns_test()
    case_func_struct_6.run_with_columns_test()
    case_func_struct_7.run_with_columns_test()
    case_func_struct_8.run_with_columns_test()
    case_format.run_with_columns_test()
    case_date_1.run_with_columns_test()
    case_date_2.run_with_columns_test()
    case_datetime_2.run_with_columns_test()


if __name__ == "__main__":
    from jsonpolars.tests import run_cov_test

    run_cov_test(__file__, "jsonpolars.expr.function", preview=False)



# -*- coding: utf-8 -*-

import typing as T
import dataclasses

import polars as pl

from ..arg import REQ, NA, rm_na, T_KWARGS
from ..base_expr import ExprEnum, BaseExpr, expr_enum_to_klass_mapping, parse_expr

if T.TYPE_CHECKING:  # pragma: no cover
    from .api import T_EXPR


def ensure_list(expr: "T_EXPR") -> pl.Expr:
    if isinstance(expr, List):
        return expr.to_polars()
    else:
        return expr.to_polars().list


@dataclasses.dataclass
class List(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/list.html
    """

    type: str = dataclasses.field(default=ExprEnum.list.value)
    expr: "T_EXPR" = dataclasses.field(default=REQ)

    @classmethod
    def from_dict(cls, dct: T_KWARGS):
        return cls(
            expr=parse_expr(dct["expr"]),
        )

    def to_polars(self) -> pl.Expr:
        return ensure_list(self.expr)


expr_enum_to_klass_mapping[ExprEnum.list.value] = List


@dataclasses.dataclass
class ListGet(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.list.get.html#polars.Expr.list.get
    """

    type: str = dataclasses.field(default=ExprEnum.list_get.value)
    expr: "T_EXPR" = dataclasses.field(default=REQ)
    index: T.Union[int, "T_EXPR"] = dataclasses.field(default=REQ)
    null_on_oob: bool = dataclasses.field(default=NA)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        req_kwargs, opt_kwargs = cls._split_req_opt(dct)
        req_kwargs["expr"] = parse_expr(req_kwargs["expr"])
        index = req_kwargs["index"]
        if isinstance(index, int):
            pass
        elif isinstance(index, dict):
            req_kwargs["index"] = parse_expr(index)
        else:  # pragma: no cover
            raise ValueError(f"Unknown index type: {index}")
        return cls(**req_kwargs, **rm_na(**opt_kwargs))

    def to_polars(self) -> pl.Expr:
        expr = ensure_list(self.expr)
        if isinstance(self.index, int):
            index = self.index
        elif isinstance(self.index, BaseExpr):
            index = self.index.to_polars()
        else:  # pragma: no cover
            raise NotImplementedError("IMPOSSIBLE TO REACH HERE!")
        return expr.get(index=index, **rm_na(null_on_oob=self.null_on_oob))


expr_enum_to_klass_mapping[ExprEnum.list_get.value] = ListGet


@dataclasses.dataclass
class ListEval(BaseExpr):
    """
    Ref: https://docs.pola.rs/api/python/stable/reference/expressions/api/polars.Expr.list.eval.html
    """

    type: str = dataclasses.field(default=ExprEnum.list_eval.value)
    expr: "T_EXPR" = dataclasses.field(default=REQ)
    expr_to_run: "T_EXPR" = dataclasses.field(default=REQ)
    parallel: bool = dataclasses.field(default=NA)

    @classmethod
    def from_dict(cls, dct: T.Dict[str, T.Any]):
        req_kwargs, opt_kwargs = cls._split_req_opt(dct)
        req_kwargs["expr"] = parse_expr(req_kwargs["expr"])
        req_kwargs["expr_to_run"] = parse_expr(req_kwargs["expr_to_run"])
        return cls(**req_kwargs, **rm_na(**opt_kwargs))

    def to_polars(self) -> pl.Expr:
        expr = ensure_list(self.expr)
        return expr.eval(self.expr_to_run.to_polars(), **rm_na(parallel=self.parallel))


expr_enum_to_klass_mapping[ExprEnum.list_eval.value] = ListEval



# -*- coding: utf-8 -*-

import copy
from jsonpolars.arg import REQ, NA, rm_na


def test_sentinel():
    for obj in [REQ, NA]:
        assert (obj == 1) is False
        assert ("a" == obj) is False
        assert (obj == [1, 2]) is False
        assert ({"a": 1} == obj) is False
        assert (obj == obj) is True
        assert (obj == copy.deepcopy(obj)) is True
        assert (copy.deepcopy(obj) == obj) is True

    assert (REQ != NA) is True
    assert (REQ == NA) is False


def test_rm_na():
    kwargs = {"a": 1, "b": NA}
    assert rm_na(**kwargs) == {"a": 1}


if __name__ == "__main__":
    from jsonpolars.tests import run_cov_test

    run_cov_test(__file__, "jsonpolars.arg", preview=False)
